(* $Id: main.ml,v 1.3 2010-07-23 15:36:03 deraugla Exp $ *)

open Printf;
open Tasks;

value version = "1.0";

value do_schedule quiet df = do {
  let ts = Sparse.make_ts df in
  let tc = ts.tasks_constraints in
  if quiet then ()
  else do {
    for i = 0 to Array.length ts.task - 1 do {
      let td = ts.task.(i) in
      printf "t%d: rel %d dur %d" (i + 1) td.release td.duration;
      match td.sub_duration with
      [ [] -> ()
      | _ -> printf "..." ];
      printf " dea %d per %d" td.deadline td.period;
      printf "\n";
      flush stdout;
      let st = Taskalg.of_task_def i td in
      printf "t%d: " (i + 1);
      printf "%s" (Taskalg.to_string st);
      printf "\n";
      flush stdout;
    }
  };
  let task_pri = do {
    let a = Array.init (Array.length ts.task) (fun i -> i) in
    Array.sort (Sparse.compare_priorities ts) a;
    a
  }
  in
  if quiet then ()
  else do {
    printf "\ntasks priorities:";
    for i = 0 to Array.length ts.task - 1 do {
      printf " t%d" (task_pri.(i) + 1);
    };
    printf "\n\n";
    flush stdout;
  };
  if Array.length task_pri = 0 then do {
    if quiet then ()
    else do {
      printf "no task\n";
      flush stdout;
    };
    0
  }
  else do {
    let accu_itask = task_pri.(0) in
    let accu_task = Taskalg.of_task_def accu_itask ts.task.(accu_itask) in
    if quiet then ()
    else do {
      printf "t%d =" (accu_itask + 1);
      printf " %s\n" (Taskalg.to_string accu_task);
    };
    let accu_task_opt =
      loop [accu_itask] 1 accu_task where rec loop stl i accu_task =
        if i >= Array.length ts.task then Some accu_task
        else do {
          let itask2 = task_pri.(i) in
          let stask2 = Taskalg.of_task_def itask2 ts.task.(itask2) in
          let stl = [itask2 :: stl] in
          if quiet then ()
          else do {
            let sl = List.rev_map (fun i -> sprintf "t%d" (i + 1)) stl in
            let s = String.concat " ⊕ " sl in
            printf "%s =" s;
            flush stdout;
          };
          match Taskalg.plus tc accu_task stask2 with
          [ Some accu_task -> do {
              if quiet then ()
              else do {
                printf " %s\n" (Taskalg.to_string accu_task);
                flush stdout;
              };
              loop stl (i+1) accu_task
            }
          | None -> do {
              if quiet then ()
              else do {
                printf "\n";
                flush stdout;
              };
              None
            } ]
        }
    in
    if quiet then ()
    else do {
      match accu_task_opt with
      [ Some accu_task -> printf "r = %s\n" (Taskalg.to_string accu_task)
      | None -> printf "not schedulable\n" ];
      flush stdout;
    };
    match accu_task_opt with
    [ Some ts -> do {
        let a = Taskalg.first_response_time ts in
        Array.sort compare a;
        printf "FRT";
        Array.iter (fun (_, wrt) -> printf " %d" wrt) a;
        printf "\n";
        flush stdout;
        let a = Taskalg.worst_response_time ts in
        Array.sort compare a;
        printf "WRT";
        Array.iter (fun (_, wrt) -> printf " %d" wrt) a;
        printf "\n";
        flush stdout;
      }
    | None -> () ];
    match accu_task_opt with
    [ Some _ -> 0
    | None -> 1 ]
  }
};

value taskadd quiet fname strm close = do {
  match
    try Left (Sparse.parse_and_check_dameid strm) with
    [ Ploc.Exc loc e -> Right (loc, e) ]
  with
  [ Left dl -> do {
      let r =
        List.fold_left (fun r (_, d) -> max r (do_schedule quiet d)) 0 dl
      in
      exit r
    }
  | Right (loc, e) -> do {
      if fname = "" then do {
        let line = Ploc.line_nb loc in
        let bp = Ploc.first_pos loc in
        let ep = Ploc.last_pos loc in
        eprintf "Line %d, characters %d-%d:\n" line bp ep;
      }
      else do {
        let (fname, line, bp, ep) = Ploc.from_file fname loc in
        eprintf "File \"%s\", line %d, characters %d-%d:\n" fname line bp ep;
      };
      match e with
      [ Stream.Error s -> eprintf "Parse error: %s\n" s
      | Plexing.Error s -> eprintf "Lexing error: %s\n" s
      | _ -> raise e ];
      flush stderr;
      exit 2;
    } ]
};

value arg_quiet = ref False;
value arg_string = ref False;
value arg_value = ref "";

value spec_list =
  Arg.align
    [("-q", Arg.Set arg_quiet,
      " Quiet: no output; exit 0 if schedulable, 1 if not");
     ("-s", Arg.Set arg_string,
      " Argument is task string instead of input file name")]
;
value anon_fun s =
  arg_value.val :=
    if not arg_string.val || arg_value.val = "" then s
    else arg_value.val ^ " " ^ s
;
value usage = "Usage: " ^ Sys.argv.(0) ^ " [options] filename-or-string";

value main () = do {
  Arg.parse spec_list anon_fun usage;
  if arg_value.val = "" then do {
    eprintf "Error: use option -help for usage\n";
    flush stderr;
    exit 2;
  }
  else do {
    let (strm, fname, close) =
      if arg_string.val then do {
(*
        if arg_quiet.val then do {
          printf "%s\n" (String.escaped arg_value.val);
          flush stdout;
        }
        else ();
*)
        let s = Plexing.eval_string Ploc.dummy arg_value.val in
        let s = s ^ "\n" in
        (Stream.of_string s, "", fun () -> ())
      }
      else do {
        let ic = open_in arg_value.val in
        let close () = close_in ic in
        (Stream.of_channel ic, arg_value.val, close)
      }
    in
    taskadd arg_quiet.val fname strm close
  }
};

main ();
