(* $Id: sparse.ml,v 1.1 2010-07-23 15:00:31 deraugla Exp $ *)

open Tasks;
open Printf;

type dameid_file =
  { d_tasks : list task_def;
    d_pree_cost : int;
    d_glob_prec : bool;
    d_graph_prec : list (list int);
    d_strict_per : bool;
    d_fully_non_pree : bool;
    d_latency_constr : list latency_constr;
    d_default_prio : option priority_policy }
;

value graph_preceding_tasks gp itask =
  let tll = List.filter (List.mem itask) gp in
  loop [] tll where rec loop rtl =
    fun
    [ [tl :: tll] ->
        loop1 rtl tl where rec loop1 rtl =
          fun
          [ [t1 :: ([t2 :: _] as tl)] ->
              if t2 = itask then loop [t1 :: rtl] tll else loop1 rtl tl
          | [_] -> loop rtl tll
          | [] -> assert False ]
    | [] -> rtl ]
;

value genuine_graph_precedence ts gp =
  List.fold_right
    (fun tl tll ->
       let tl =
         loop tl where rec loop =
           fun
           [ [t1; t2 :: tl] ->
               if t1 >= Array.length ts.task then loop [t2 :: tl]
               else if t2 >= Array.length ts.task then loop [t1 :: tl]
(*
               else if compare_priorities ts t1 t2 < 0 then
               else if ts.task.(t1).period <= ts.task.(t2).period then
                 [t1 :: loop [t2 :: tl]]
               else loop [t1 :: tl]
*)
               else [t1 :: loop [t2 :: tl]]
(**)
           | tl -> tl ]
       in
       match tl with
       [ [] | [_] -> tll
       | _ -> [tl :: tll] ])
    gp []
;

value rate_monotonic_compare itask1 itask2 td1 td2 =
  if td1.period < td2.period then -1
  else if td1.period > td2.period then 1
  else if td1.deadline < td2.deadline then -1
  else if td1.deadline > td2.deadline then 1
  else itask1 - itask2
;

value deadline_monotonic_compare itask1 itask2 td1 td2 =
  if td1.deadline < td2.deadline then -1
  else if td1.deadline > td2.deadline then 1
  else if td1.period < td2.period then -1
  else if td1.period > td2.period then 1
  else itask1 - itask2
;

value compare_priorities ts itask1 itask2 =
  let tc = ts.tasks_constraints in
  let td1 = ts.task.(itask1) in
  let td2 = ts.task.(itask2) in
  if tc.global_precedence then
    (* Always rate monotonic *)
    rate_monotonic_compare itask1 itask2 td1 td2
  else
    let tll =
      List.filter (fun tl -> List.mem itask1 tl && List.mem itask2 tl)
        tc.graph_precedence
    in
    match tll with
    [ [tl :: _] ->
        (* Same graph => rate monotonic *)
        if td1.period < td2.period then -1
        else if td2.period < td1.period then 1
        else
          loop tl where rec loop =
            fun
            [ [t :: tl] ->
                if t = itask1 then -1
                else if t = itask2 then 1
                else loop tl
            | [] -> assert False ]
    | [] ->
        (* No graph or different graphs => menu *)
        match ts.priority_policy with
        [ PP_user t ->
            compare t.(itask1) t.(itask2)
        | PP_deadline_monotonic ->
            deadline_monotonic_compare itask1 itask2 td1 td2
        | PP_rate_monotonic ->
            rate_monotonic_compare itask1 itask2 td1 td2
        | PP_edf ->
            compare itask1 itask2
        | PP_audsley {ai_tab = t} ->
            compare t.(itask1) t.(itask2) ] ]
;

value set_task_order ts = do {
  let tc = ts.tasks_constraints in
  let task_order =
    let list =
      let list = Sutil.rho (Array.length ts.task) in
      match (tc.global_precedence, tc.graph_precedence) with
      [ (True, _) | (_, [_ :: _]) ->
          List.sort (fun it1 it2 -> compare_priorities ts it1 it2) list
      | _ ->
          list ]
    in
    Array.of_list list
  in
  for i = 0 to Array.length ts.task - 1 do {
    ts.task_order.(i) := task_order.(i);
  };
};

value make_ts d = do {
  let tdl =
    if d.d_glob_prec then
      List.map (fun td -> {(td) with release = 0}) d.d_tasks
    else
      d.d_tasks
  in
  let ntasks = List.length tdl in
  let task_order = Array.of_list (Sutil.rho ntasks) in
  let (min_rel_time, max_rel_time) =
    List.fold_left
      (fun (min_rel_time, max_rel_time) td ->
         (min min_rel_time td.release, max max_rel_time td.release))
      (max_int, min_int) tdl
  in
  let rmax = max_rel_time - min_rel_time in
  let lcm = Sutil.lcm_of_list (List.map (fun td -> td.period) tdl) in
  let schedule_len = if rmax = 0 then lcm else rmax + 2 * lcm in
  let disp_schedule = Sarray.make schedule_len None in
  let comp_schedule = Sarray.make schedule_len None in
  let priority_policy =
    match d.d_default_prio with
    [ Some p -> p
    | None -> PP_deadline_monotonic ]
  in
  let tc =
    {global_precedence = d.d_glob_prec; graph_precedence = [];
     preemption_cost = d.d_pree_cost}
  in
  let ts =
    {task = Array.of_list tdl; task_order = task_order;
     priority_policy = priority_policy;
     user_graph_precedence = d.d_graph_prec; tasks_constraints = tc;
     strict_periodicity = d.d_strict_per;
     fully_non_preemptive = d.d_fully_non_pree;
     latency_constr = d.d_latency_constr;

     schedule_len = schedule_len;
     disp_schedule = disp_schedule; comp_schedule = comp_schedule;
     rmax = rmax; lcm = lcm; min_rel_time = min_rel_time;

     schedule_status = SS_running; repetition = None;
     curr_ord_ind = 0; curr_inst = 0;
     do_refresh = False}
  in
  tc.graph_precedence := genuine_graph_precedence ts ts.user_graph_precedence;
  set_task_order ts;
  ts
};

value first_duration td =
  List.fold_left (fun accu sd -> accu - sd.sub_dur) td.duration
    td.sub_duration
;

(* TODO: other tests to be added than just first_duration *)
value check_task td =
  if first_duration td < 0 then False else True
;

value check_dameid dm =
  loop dm.d_tasks where rec loop =
    fun
    [ [td :: tdl] -> if check_task td then loop tdl else False
    | [] -> True ]
;

#load "pa_extend.cmo";
#load "pa_lexer.cmo";

value rec number = lexer [ '0'-'9' number! | ];
value rec ident = lexer [ [ 'a'-'z' | 'A'-'Z' ] ident! | ];
value rec comment = lexer [ '\n'/ | _/ comment! ];

value rec next_token =
  lexer
  [ ' '/ next_token!
  | '#'/ comment! next_token!
  | '\n' -> ("EOL", "")
  | '/' -> ("", "/")
  | [ 'a'-'z' | 'A'-'Z' ] ident! -> ("", $buf)
  | '0'-'9' number! -> ("INT", $buf)
  | '-' '0'-'9' number! -> ("INT", $buf)
  | '-' '>' -> ("", $buf)
  | '(' -> ("", $buf)
  | ')' -> ("", $buf)
  | ':' -> ("", $buf)
  | '<' -> ("", $buf)
  | ',' -> ("", $buf)
  | _ -> raise (Plexing.Error (sprintf "illegal character '%s'" $buf))
  | -> ("EOI", "") ]
;

value lexer_fun (strm, line_nb, bol_pos) = do {
  let bp = Stream.count strm in
  let r = next_token $empty strm in
(*
  eprintf "r = \"%s\",\"%s\"\n" (fst r) (snd r); flush stderr;
*)
  let ep = Stream.count strm  in
  let loc = Ploc.make line_nb.val bol_pos.val (bp, ep) in
  (r, loc)
};

value lex =
  { Plexing.tok_func = Plexing.lexer_func_of_parser lexer_fun;
    Plexing.tok_using _ = ();
    Plexing.tok_removing _ = ();
    Plexing.tok_match = Plexing.default_match;
    Plexing.tok_text = Plexing.lexer_text;
    Plexing.tok_comm = None }
;

value gram = Grammar.gcreate lex;
value list_scenario_eoi = Grammar.Entry.create gram "list scenario";
value list_latency_eoi = Grammar.Entry.create gram "list latency";
value graph_prec_eoi = Grammar.Entry.create gram "precedence";
value user_priority_eoi = Grammar.Entry.create gram "priority";

EXTEND
  GLOBAL: list_scenario_eoi list_latency_eoi graph_prec_eoi user_priority_eoi;
  list_scenario_eoi:
    [ [ el = list_scenario; EOI -> el ] ]
  ;
  list_scenario:
    [ [ s = scenario; "EOD"; eol; sl = list_scenario -> [s :: sl]
      | s = scenario; "EOD"; eol -> [s]
      | s = scenario -> [s] ] ]
  ;
  scenario:
    [ [ pree_cost = int; eol; tl = LIST0 task;
        glob_prec = FLAG [ "prec"; EOL ];
        graph_prec = graph_prec_constr;
        strict_per = FLAG [ "strp"; eol ];
        fully_non_pree = FLAG [ "fnp"; eol ];
        latency_constr = latency_constr;
        default_prio = OPT default_prio ->
          let r =
            {d_tasks = tl; d_pree_cost = pree_cost; d_glob_prec = glob_prec;
             d_graph_prec = graph_prec; d_strict_per = strict_per;
             d_fully_non_pree = fully_non_pree;
             d_latency_constr = latency_constr;
             d_default_prio = default_prio}
          in
          (loc, r) ] ]
  ;
  task:
    [ [ rel = int; dur = int; sub_dur = LIST0 sub_duration; dea = int;
        (per, npree, prco) = int_and_npre; eol ->
          let per = if per = 0 then dea else per in
          let dea = min dea per in
(*
let _ = do { eprintf "rel %d dur %d dea %d per %d\n" rel dur dea per; flush stderr } in
*)
          {release = rel; duration = dur; sub_duration = sub_dur;
           deadline = dea; period = per;
           user_release = rel; non_preemptive = npree;
           strict_per = False; pree_cost = prco} ] ]
  ;
  sub_duration:
    [ [ "("; rel = int; dur = int; non_pree = FLAG "np"; ")" ->
          {sub_rel = rel; sub_dur = dur; sub_non_pree = non_pree;
           sub_str_per = False} ] ]
  ;
  int_and_npre:
    [ [ n = int; npre = FLAG "np"; pcro = preemption_cost -> (n, npre, pcro)
      | npre = FLAG "np"; pcro = preemption_cost -> (0, npre, pcro)
      | -> (0, False, 0) ] ]
  ;
  preemption_cost:
    [ [ "pc"; n = int -> n
      | -> 0 ] ]
  ;
  graph_prec_constr:
    [ [ "prec"; ":"; g = LIST0 task_prec; eol -> g
      | -> [] ] ]
  ;
  graph_prec_eoi:
    [ [ g = LIST0 task_prec; EOI -> g ] ]
  ;
  task_prec:
    [ [ t = task_name; tl = LIST1 [ "<"; t = task_name -> t ] -> [t :: tl] ] ]
  ;
  latency_constr:
    [ [ "lat"; ":"; ll = LIST1 latency; eol -> ll
      | -> [] ] ]
  ;
  list_latency_eoi:
    [ [ ll = LIST0 latency; EOI -> ll ] ]
  ;
  latency:
    [ [ nt1 = task_name; "("; inst1 = int; ")"; "->"; nt2 = task_name; "(";
       inst2 = int; ")"; "<"; mlat = int ->
         {l_task1 = nt1; l_inst1 = inst1; l_task2 = nt2; l_inst2 = inst2;
          l_max = mlat; l_status = LS_unchecked} ] ]
  ;
  task_name:
    [ [ "t"; n = int -> n - 1 ] ]
  ;
  default_prio:
    [ [ "prio"; ":"; a = prio; eol -> a ] ]
  ;
  prio:
    [ [ "EDF" -> PP_edf ] ]
  ;
  int:
    [ [ n = INT -> int_of_string n ] ]
  ;
  eol:
    [ [ "/" | LIST1 EOL -> () ] ]
  ;
  user_priority_eoi:
    [ [ ll = LIST0 task_name SEP ","; EOI -> ll ] ]
  ;
END;

value parse_and_check_dameid strm = do {
  let dl = Grammar.Entry.parse list_scenario_eoi strm in
  List.iter
    (fun (loc, d) -> do {
       if check_dameid d then ()
       else Ploc.raise loc (Failure "inconsistency in task list")
     })
    dl;
  dl
};

value parse_handled entry entry_name strm =
  try Some (Grammar.Entry.parse entry strm) with
  [ Ploc.Exc loc e -> do {
      let bp = Ploc.first_pos loc in
      let ep = Ploc.last_pos loc in
      eprintf "%s, characters %d-%d:\n" entry_name bp ep;
      match e with
      [ Stream.Error s -> eprintf "Parse error: %s\n" s
      | Plexing.Error s -> eprintf "Lexing error: %s\n" s
      | _ -> eprintf "%s\n" (Printexc.to_string e) ];
      flush stderr;
      None
    } ]
;

value parse_latency_list_eoi = parse_handled list_latency_eoi "Latency";
value parse_graph_prec_eoi = parse_handled graph_prec_eoi "Precedence";
value parse_user_priority_eoi = parse_handled user_priority_eoi "Priority";
